import kfactory as kf
import pytest
from conftest import Layers


@pytest.mark.parametrize(
    "center",
    [
        None,
        kf.kdb.Point(0, 0),
        kf.kdb.Point(500, 0),
        kf.kdb.Point(500, 500),
        kf.kdb.Point(1000, 1000),
    ],
)
def test_rotation(
    center: kf.kdb.Point | None, straight: kf.KCell, LAYER: Layers
) -> None:
    c = kf.KCell()

    wg1 = c << straight
    wg2 = c << straight

    wg2.rotate(2, center=center)

    if center:
        c.shapes(c.kcl.find_layer(LAYER.WGCLAD)).insert(
            kf.kdb.Box(10).transformed(kf.kdb.Trans(center.to_v()))
        )

    c.add_ports(wg1.ports)
    c.add_ports(wg2.ports)


@pytest.mark.parametrize(
    "center",
    [
        None,
        kf.kdb.DPoint(0, 0),
        kf.kdb.DPoint(0.5, 0),
        kf.kdb.DPoint(0.5, 0.5),
        kf.kdb.DPoint(1, 1),
    ],
)
def test_drotation(
    center: kf.kdb.DPoint | None, straight: kf.KCell, LAYER: Layers
) -> None:
    c = kf.KCell()

    wg1 = c << straight
    wg2 = c << straight

    wg2.drotate(30, center=center)

    if center:
        c.shapes(c.kcl.find_layer(LAYER.WGCLAD)).insert(
            kf.kdb.DBox(0.01).transformed(kf.kdb.DCplxTrans(center.to_v()))
        )

    c.add_ports(wg1.ports)
    c.add_ports(wg2.ports)


@pytest.mark.parametrize(
    "from_name,use_mirror,apply_mirror,expected_transformation",
    [
        (True, True, True, kf.kdb.Trans(1, False, 11_000, -10_000)),
        (True, True, False, kf.kdb.Trans(1, False, 11_000, -10_000)),
        (True, False, True, kf.kdb.Trans(3, True, 11_000, 10_000)),
        (True, False, False, kf.kdb.Trans(1, False, 11_000, -10_000)),
        (False, True, True, kf.kdb.Trans(1, False, 11_000, -10_000)),
        (False, True, False, kf.kdb.Trans(1, False, 11_000, -10_000)),
        (False, False, True, kf.kdb.Trans(3, True, 11_000, 10_000)),
        (False, False, False, kf.kdb.Trans(1, False, 11_000, -10_000)),
    ],
)
def test_connection_flags(
    straight: kf.KCell,
    bend90: kf.KCell,
    from_name: bool,
    use_mirror: bool,
    apply_mirror: bool,
    expected_transformation: kf.kdb.Trans,
) -> None:
    """Tests all the (relevant) connection flags."""
    c = kf.KCell(f"{from_name=}_{use_mirror=}_{apply_mirror=}")
    i1 = c << straight
    i2 = c << bend90

    if apply_mirror:
        i2.mirror()

    port: str | kf.Port = "o2" if from_name else i2.ports["o2"]

    i2.connect(port, i1, "o2", use_mirror=use_mirror)

    assert i2.trans == expected_transformation



"""Circular bends.

A circular bend has a constant radius.
"""

from collections.abc import Callable
from typing import Any, Protocol

import numpy as np

from .. import kdb, kf_types
from ..conf import logger
from ..enclosure import LayerEnclosure, extrude_path
from ..kcell import Info, KCell, KCLayout, MetaData

__all__ = ["bend_circular_factory"]


class BendCircularKCell(Protocol):
    def __call__(
        self,
        width: kf_types.um,
        radius: kf_types.um,
        layer: kdb.LayerInfo,
        enclosure: LayerEnclosure | None = None,
        angle: kf_types.deg = 90,
        angle_step: kf_types.deg = 1,
    ) -> KCell:
        """Circular radius bend [um].

        Args:
            width: Width of the core. [um]
            radius: Radius of the backbone. [um]
            layer: Layer index of the target layer.
            enclosure: Optional enclosure.
            angle: Angle amount of the bend.
            angle_step: Angle amount per backbone point of the bend.
        """
        ...


def bend_circular_factory(
    kcl: KCLayout,
    additional_info: Callable[
        ...,
        dict[str, MetaData],
    ]
    | dict[str, MetaData]
    | None = None,
    basename: str | None = None,
    snap_ports: bool = False,
    **cell_kwargs: Any,
) -> BendCircularKCell:
    """Returns a function generating circular bends.

    Args:
        kcl: The KCLayout which will be owned
        additional_info: Add additional key/values to the
            [`KCell.info`][kfactory.kcell.KCell.info]. Can be a static dict
            mapping info name to info value. Or can a callable which takes the straight
            functions' parameters as kwargs and returns a dict with the mapping.
        basename: Overwrite the prefix of the resulting KCell's name. By default
            the KCell will be named 'straight_dbu[...]'.
        snap_ports: Whether to snap ports to grid.
        cell_kwargs: Additional arguments passed as `@kcl.cell(**cell_kwargs)`.
    """
    if callable(additional_info) and additional_info is not None:
        _additional_info_func: Callable[
            ...,
            dict[str, MetaData],
        ] = additional_info
        _additional_info: dict[str, MetaData] = {}
    else:

        def additional_info_func(
            **kwargs: Any,
        ) -> dict[str, MetaData]:
            return {}

        _additional_info_func = additional_info_func
        _additional_info = additional_info or {}

    @kcl.cell(
        basename=basename,
        snap_ports=snap_ports,
        **cell_kwargs,
    )
    def bend_circular(
        width: kf_types.um,
        radius: kf_types.um,
        layer: kdb.LayerInfo,
        enclosure: LayerEnclosure | None = None,
        angle: kf_types.deg = 90,
        angle_step: kf_types.deg = 1,
    ) -> KCell:
        """Circular radius bend [um].

        Args:
            width: Width of the core. [um]
            radius: Radius of the backbone. [um]
            layer: Layer index of the target layer.
            enclosure: Optional enclosure.
            angle: Angle amount of the bend.
            angle_step: Angle amount per backbone point of the bend.
        """
        c = kcl.kcell()
        r = radius

        if angle < 0:
            logger.critical(
                f"Negative angles are not allowed {angle} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            angle = -angle
        if width < 0:
            logger.critical(
                f"Negative widths are not allowed {width} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            width = -width

        backbone = [
            kdb.DPoint(x, y)
            for x, y in [
                [
                    np.sin(_angle / 180 * np.pi) * r,
                    (-np.cos(_angle / 180 * np.pi) + 1) * r,
                ]
                for _angle in np.linspace(
                    0, angle, int(angle // angle_step + 0.5), endpoint=True
                )
            ]
        ]

        center_path = extrude_path(
            target=c,
            layer=layer,
            path=backbone,
            width=width,
            enclosure=enclosure,
            start_angle=0,
            end_angle=angle,
        )

        c.create_port(
            trans=kdb.Trans(2, False, 0, 0),
            width=int(width / c.kcl.dbu),
            layer=c.kcl.find_layer(layer),
        )
        c.create_port(
            dcplx_trans=kdb.DCplxTrans(1, angle, False, backbone[-1].to_v()),
            dwidth=width,
            layer=c.kcl.find_layer(layer),
        )
        c.auto_rename_ports()
        c.boundary = center_path
        _info: dict[str, MetaData] = {}
        _info.update(
            _additional_info_func(
                width=width,
                radius=radius,
                layer=layer,
                enclosure=enclosure,
                angle=angle,
                angle_step=angle_step,
            )
        )
        _info.update(_additional_info)
        c.info = Info(**_info)
        return c

    return bend_circular



"""Placer of bends/straights from a route."""

import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, TypeVar

from ruamel.yaml import YAML
from ruamel.yaml.constructor import SafeConstructor

from .enclosure import LayerEnclosure
from .kcell import KCell, KCLayout, Port, Ports
from .kcell import kcl as stdkcl

__all__ = ["cells_to_yaml", "cells_from_yaml"]

PathLike = TypeVar("PathLike", str, Path, None)


def cells_to_yaml(output: PathLike, cells: list[KCell] | KCell) -> None:
    """Convert cell(s) to a yaml representations.

    Args:
        output: A stream or string of a path where to dump the yaml. Can also be
            set to sys.stdout
        cells: A single [KCell][kfactory.kcell.KCell] or a list of them.


    Returns:
        yaml dump
    """
    cells = list(cells)
    cells.sort(key=lambda c: c.hierarchy_levels())
    yaml = YAML()
    yaml.register_class(KCell)
    yaml.register_class(Port)
    yaml.register_class(Ports)
    yaml.indent(sequence=4, offset=2)
    yaml.dump(cells, output)


def get_yaml_obj() -> YAML:
    """New global yaml object."""
    return YAML()


def register_classes(
    yaml: YAML,
    kcl: KCLayout = stdkcl,
    additional_classes: list[object] | None = None,
    verbose: bool = False,
) -> None:
    """Register a new KCell class compatible with ruamel yaml."""

    class ModKCell(KCell):
        def __init__(self, name: str | None = None, library: KCLayout = kcl):
            KCell.__init__(self, name, library)

        @classmethod
        def from_yaml(cls, constructor, node):  # type: ignore[no-untyped-def]
            return super().from_yaml(constructor, node, verbose=verbose)

    yaml.register_class(ModKCell)
    yaml.register_class(Port)
    yaml.register_class(Ports)
    yaml.register_class(LayerEnclosure)

    if additional_classes is not None:
        for c in additional_classes:
            yaml.register_class(c)


def cells_from_yaml(
    inp: Path,
    kcl: KCLayout = stdkcl,
    additional_classes: list[object] | None = None,
    verbose: bool = False,
) -> None:
    """Recreate cells from a yaml file.

    Args:
        inp: Input file path.
        kcl: KCLayout to load the cells into.
        additional_classes: Additional yaml classes that should be registered.
            This is used for example to enable loading additional yaml files etc.
        verbose: Print more verbose errors etc.
    """
    yaml = get_yaml_obj()
    yaml.register_class(
        include_from_loader(inp.parent, kcl, additional_classes, verbose)
    )

    register_classes(
        yaml,
        kcl,
        additional_classes,
        verbose,
    )
    yaml.load(inp)


def exploded_yaml(
    inp: os.PathLike[Any],
    library: KCLayout = stdkcl,
    additional_classes: list[object] | None = None,
    verbose: bool = False,
) -> Any:
    """Expanded yaml.

    Expand cross-references. Same syntax as :py:func:~`cells_from_yaml`
    """
    yaml = YAML(pure=True)

    class ModKCell(KCell):
        def __init__(self, name: str | None = None, library: KCLayout = library):
            KCell.__init__(self, name, library)

        @classmethod
        def from_yaml(cls, constructor, node):  # type: ignore[no-untyped-def]
            super().from_yaml(constructor, node, verbose=verbose)

    return yaml.dump(yaml.load(inp), sys.stdout)


def include_from_loader(
    folder: Path,
    library: KCLayout,
    additional_classes: list[object] | None,
    verbose: bool,
) -> Any:
    """Expand ruamel to support the `!include` keyword."""

    @dataclass
    class Include:
        filename: str
        yaml_tag: str = "!include"

        @classmethod
        def from_yaml(cls, constructor, node):  # type: ignore[no-untyped-def]
            d = SafeConstructor.construct_mapping(constructor, node)

            f = Path(d["filename"])
            if f.is_absolute():
                cells_from_yaml(f, library, additional_classes, verbose)
            else:
                cells_from_yaml(
                    (folder / f).resolve(), library, additional_classes, verbose
                )

    return Include



"""Create 1D or 2D (flex) grids in KCells."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Literal, cast

from . import kdb
from .kcell import Instance, KCell


def grid_dbu(
    target: KCell,
    kcells: Sequence[KCell | None] | Sequence[Sequence[KCell | None]],
    spacing: int | tuple[int, int],
    target_trans: kdb.Trans = kdb.Trans(),
    shape: tuple[int, int] | None = None,
    align_x: Literal["origin", "xmin", "xmax", "center"] = "center",
    align_y: Literal["origin", "ymin", "ymax", "center"] = "center",
    rotation: Literal[0, 1, 2, 3] = 0,
    mirror: bool = False,
) -> list[list[Instance | None]]:
    """Create a grid of instances.

    A grid uses the bounding box of the biggest width and biggest height of any bounding
    boxes inserted into the grid.
    to this bounding box a spacing is applied in x and y.

    ```
                          spacing[0] or spacing
                         ◄─►
      ┌──────────────────┐ ┌────┬─────┬─────┐ ┌────────────────┐ ┌──────────────────┐ ▲
      │                  │ │    │     │     │ │                │ │                  │ │
      │   ┌────┐         │ │    │     │     │ │                │ │      ┌──────┐    │ │
      │   │    │         │ │    │     │     │ │                │ │      │      │    │ │
      │   │    │         │ │    │ big │     │ │    ┌───────┐   │ │      │      │    │ │
      │   │    │         │ │    │ comp│     │ │    │       │   │ │      │      │    │ │
      │   │    │         │ │    │ y   │     │ │    │       │   │ │      │      │    │ │
      │   │    │         │ │    │     │     │ │    │       │   │ │      │     max bbox y
      │   └────┘         │ │    │     │     │ │    │       │   │ │      │      │    │ │
      │                  │ │    │     │     │ │    │       │   │ │      │      │    │ │
      │                  │ │    │     │     │ │    │       │   │ │      └──────┘    │ │
      │                  │ │    │     │     │ │    │       │   │ │                  │ │
      │                  │ │    │     │     │ │    └───────┘   │ │                  │ │
      │                  │ │    │     │     │ │                │ │                  │ │
     ▲└──────────────────┘ └────┴─────┴─────┘ └────────────────┘ └──────────────────┘ ▼
     │spacing[1] or spacing
     ▼┌──────────────────┐ ┌────────────────┐ ┌────────────────┐ ┌──────────────────┐
      │                  │ │                │ │                │ │                  │
      │                  │ │                │ │ ┌────┐         │ │  ┌───┐           │
      ├──────────────────┤ │ ┌───────────┐  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      │   big comp x     │ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      ├──────────────────┤ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ └───────────┘  │ │ └────┘         │ │  └───┘           │
      │                  │ │                │ │                │ │                  │
      └──────────────────┘ └────────────────┘ └────────────────┘ └──────────────────┘

                                                                 ◄──────────────────►
                                                                    max bbox x
    ```

    Args:
        target: Target KCell.
        kcells: Sequence or sequence of sequence of KCells to add to the grid
        spacing: Value or tuple of value (different x/y) for spacing of the grid. [dbu]
        target_trans: Apply a transformation to the whole grid before placing it.
        shape: Respace the input of kcells into an array and fill as many positions
            (first x then y).
        align_x: Align all the instance on the x-coordinate.
        align_y: Align all the instance on the y-coordinate.
        rotation: Apply a rotation to each kcells instance before adding it to the grid.
        mirror: Mirror the instances before placing them in the grid

    """
    if isinstance(spacing, tuple):
        spacing_x, spacing_y = spacing
    else:
        spacing_x = spacing
        spacing_y = spacing

    insts: list[list[Instance | None]]
    kcell_array: Sequence[Sequence[KCell]]

    if shape is None:
        if isinstance(kcells[0], KCell):
            kcell_array = [list(kcells)]  # type:ignore[arg-type]
        else:
            kcell_array = kcells  # type: ignore[assignment]

        x0 = 0
        y0 = 0

        insts = [
            [
                target.create_inst(kcell, kdb.Trans(rotation, mirror, 0, 0))
                for kcell in array
            ]
            for array in kcell_array
        ]
        bboxes = [
            [None if inst is None else inst.bbox() for inst in array] for array in insts
        ]
        w = max(
            max(0 if bbox is None else bbox.width() + spacing_x for bbox in box_array)
            for box_array in bboxes
        )
        h = max(
            max(0 if bbox is None else bbox.height() + spacing_y for bbox in box_array)
            for box_array in bboxes
        )
        for array, bbox_array in zip(insts, bboxes):
            y0 += h - h // 2
            for bbox, inst in zip(bbox_array, array):
                x0 += w - w // 2
                if bbox is not None and inst is not None:
                    match align_x:
                        case "xmin":
                            x = -bbox.left
                        case "xmax":
                            x = -bbox.right
                        case "center":
                            x = -bbox.center().x
                        case _:
                            x = 0
                    match align_y:
                        case "ymin":
                            y = -bbox.bottom
                        case "ymax":
                            y = -bbox.top
                        case "center":
                            y = -bbox.center().y
                        case _:
                            y = 0
                    at = kdb.Trans(x0 + x, y0 + y)

                    inst.transform(target_trans * at)
                x0 += w // 2
            y0 += h // 2
            x0 = 0
        return insts
    else:
        _kcells: list[KCell | None]
        if isinstance(kcells[0], KCell):
            _kcells = kcells  # type:ignore[assignment]
        else:
            _kcells = [kcell for array in kcells for kcell in array]  # type: ignore[union-attr]

        if len(_kcells) > shape[0] * shape[1]:
            raise ValueError(
                f"Shape container size {shape[0] * shape[1]=!r} must be bigger "
                f"than the number of kcells {len(_kcells)}"
            )

        x0 = 0
        y0 = 0

        _insts = [
            None
            if kcell is None
            else target.create_inst(kcell, kdb.Trans(rotation, mirror, 0, 0))
            for kcell in _kcells
        ]

        insts = []
        for _ in range(shape[0]):
            insts.append([None] * shape[1])

        shape_bboxes = [None if inst is None else inst.bbox() for inst in _insts]
        shape_bboxes_heights = [
            0 if box is None else box.height() for box in shape_bboxes
        ]
        shape_bboxes_widths = [
            0 if box is None else box.width() for box in shape_bboxes
        ]
        w = max(shape_bboxes_widths) + spacing_x
        h = max(shape_bboxes_heights) + spacing_y
        for i, (inst, bbox) in enumerate(zip(_insts, shape_bboxes)):
            i_x = i % shape[1]
            i_y = i // shape[1]
            insts[i_y][i_x] = inst
            if i_x == 0:
                y0 += h - h // 2
                x0 = 0
            else:
                x0 += w - w // 2

            if bbox is not None and inst is not None:
                match align_x:
                    case "xmin":
                        x = -bbox.left
                    case "xmax":
                        x = -bbox.right
                    case "center":
                        x = -bbox.center().x
                    case _:
                        x = 0
                match align_y:
                    case "ymin":
                        y = -bbox.bottom
                    case "ymax":
                        y = -bbox.top
                    case "center":
                        y = -bbox.center().y
                    case _:
                        y = 0
                at = kdb.Trans(x0 + x, y0 + y)

                inst.transform(target_trans * at)
            if i_x == shape[1] - 1:
                y0 += h // 2
                x0 = 0
            else:
                x0 += w // 2
        return insts


def flexgrid_dbu(
    target: KCell,
    kcells: Sequence[KCell | None] | Sequence[Sequence[KCell | None]],
    spacing: int | tuple[int, int],
    target_trans: kdb.Trans = kdb.Trans(),
    shape: tuple[int, int] | None = None,
    align_x: Literal["origin", "xmin", "xmax", "center"] = "center",
    align_y: Literal["origin", "ymin", "ymax", "center"] = "center",
    rotation: Literal[0, 1, 2, 3] = 0,
    mirror: bool = False,
) -> list[list[Instance | None]]:
    """Create a grid of instances.

    A grid uses the bounding box of the biggest width per column and biggest height per
    row of any bounding boxes inserted into the grid.
    To this bounding box a spacing is applied in x and y.

    ```
                         spacing[0] or spacing
                        ◄─►
     ┌──────────────────┐ ┌──┬─────┬──┐ ┌──────────┐ ┌──────────┐▲
     │                  │ │  │     │  │ │          │ │          ││
     │   ┌────┐         │ │  │     │  │ │          │ │   ┌──────┼│
     │   │    │         │ │  │     │  │ │          │ │   │      ││
     │   │    │         │ │  │ big │  │ │  ┌───────┤ │   │      ││
     │   │ ▼  │         │ │  │ comp│  │ │  │       │ │   │      ││
     │   │    │         │ │  │ y   │  │ │  │       │ │   │  ▼   ││
     │   │    │         │ │  │     │  │ │  │  ▼    │ │   │      ││ y[1]
     │   └────┘         │ │  │ ▼   │  │ │  │       │ │   │      ││
     │                  │ │  │     │  │ │  │       │ │   │      ││
     │                  │ │  │     │  │ │  │       │ │   └──────┼│
     │                  │ │  │     │  │ │  │       │ │          ││
     │                  │ │  │     │  │ │  └───────┤ │          ││
     │                  │ │  │     │  │ │          │ │          ││
    ▲└──────────────────┘ └──┴─────┴──┘ └──────────┘ └──────────┘▼
    │spacing[1] or spacing
    ▼┌──────────────────┐ ┌───────────┐ ┌────┬─────┐ ┌───┬──────┐▲
     ├──────────────────┤ ├───────────┤ │    │     │ │   │      ││
     │                  │ │           │ │    │     │ │   │      ││
     │              ▼   │ │           │ │    │     │ │   │      ││
     │   big comp x     │ │           │ │    │     │ │   │      ││
     │                  │ │    ▼      │ │  ▼ │     │ │ ▼ │      ││ y[0]
     ├──────────────────┤ │           │ │    │     │ │   │      ││
     │                  │ │           │ │    │     │ │   │      ││
     │                  │ │           │ │    │     │ │   │      ││
     │                  │ │           │ │    │     │ │   │      ││
     └──────────────────┴─┴───────────┴─┴────┼─────┴─┴───┼──────┘►
     ►──────────────────► ►───────────► ►──────────► ►──────────►
             x[0]               x[1]        x[2]         x[3]
    ```

    Args:
        target: Target KCell.
        kcells: Sequence or sequence of sequence of KCells to add to the grid
        spacing: Value or tuple of value (different x/y) for spacing of the grid. [dbu]
        target_trans: Apply a transformation to the whole grid before placing it.
        shape: Respace the input of kcells into an array and fill as many positions
            (first x then y).
        align_x: Align all the instance on the x-coordinate.
        align_y: Align all the instance on the y-coordinate.
        rotation: Apply a rotation to each kcells instance before adding it to the grid.
        mirror: Mirror the instances before placing them in the grid

    """
    if isinstance(spacing, tuple):
        spacing_x, spacing_y = spacing
    else:
        spacing_x = spacing
        spacing_y = spacing

    insts: list[list[Instance | None]]
    kcell_array: Sequence[Sequence[KCell]]

    if shape is None:
        if isinstance(kcells[0], KCell):
            kcell_array = cast(Sequence[list[KCell]], [list(kcells)])
        else:
            kcell_array = cast(Sequence[Sequence[KCell]], kcells)

        x0 = 0
        y0 = 0

        insts = [
            [
                None
                if kcell is None
                else target.create_inst(kcell, kdb.Trans(rotation, mirror, 0, 0))
                for kcell in array
            ]
            for array in kcell_array
        ]
        bboxes = [
            [None if inst is None else inst.bbox() for inst in array] for array in insts
        ]
        xmin: dict[int, int] = {}
        ymin: dict[int, int] = {}
        ymax: dict[int, int] = {}
        xmax: dict[int, int] = {}
        for i_y, (array, box_array) in enumerate(zip(insts, bboxes)):
            for i_x, (inst, bbox) in enumerate(zip(array, box_array)):
                if inst is not None and bbox is not None:
                    if inst is not None and bbox is not None:
                        match align_x:
                            case "xmin":
                                x = -bbox.left
                            case "xmax":
                                x = -bbox.right
                            case "center":
                                x = -bbox.center().x
                            case _:
                                x = 0
                        match align_y:
                            case "ymin":
                                y = -bbox.bottom
                            case "ymax":
                                y = -bbox.top
                            case "center":
                                y = -bbox.center().y
                            case _:
                                y = 0
                        at = kdb.Trans(x, y)
                        inst.trans = at * inst.trans
                        bbox = inst.bbox()
                        xmin[i_x] = min(
                            xmin.get(i_x, None) or bbox.left, bbox.left - spacing_x
                        )
                        xmax[i_x] = max(xmax.get(i_x, None) or bbox.right, bbox.right)
                        ymin[i_y] = min(
                            ymin.get(i_y, None) or bbox.bottom, bbox.bottom - spacing_y
                        )
                        ymax[i_y] = max(ymax.get(i_y, None) or bbox.top, bbox.top)

        for i_y, (array, bbox_array) in enumerate(zip(insts, bboxes)):
            y0 -= ymin.get(i_y, 0)
            for i_x, (bbox, inst) in enumerate(zip(bbox_array, array)):
                x0 -= xmin.get(i_x, 0)
                if inst is not None and bbox is not None:
                    at = kdb.Trans(x0, y0)
                    inst.transform(target_trans * at)
                x0 += xmax.get(i_x, 0)
            y0 += ymax.get(i_y, 0)
            x0 = 0
        return insts
    else:
        _kcells: list[KCell | None]
        if isinstance(kcells[0], KCell):
            _kcells = kcells  # type:ignore[assignment]
        else:
            _kcells = [kcell for array in kcells for kcell in array]  # type: ignore[union-attr]

        if len(_kcells) > shape[0] * shape[1]:
            raise ValueError(
                f"Shape container size {shape[0] * shape[1]=} must be bigger "
                f"than the number of kcells {len(_kcells)}"
            )

        x0 = 0
        y0 = 0

        _insts = [
            None
            if kcell is None
            else target.create_inst(kcell, kdb.Trans(rotation, mirror, 0, 0))
            for kcell in _kcells
        ]

        xmin = {}
        ymin = {}
        ymax = {}
        xmax = {}
        for i, inst in enumerate(_insts):
            i_x = i % shape[1]
            i_y = i // shape[1]

            if inst is not None:
                bbox = inst.bbox()
                match align_x:
                    case "xmin":
                        x = -bbox.left
                    case "xmax":
                        x = -bbox.right
                    case "center":
                        x = -bbox.center().x
                    case _:
                        x = 0
                match align_y:
                    case "ymin":
                        y = -bbox.bottom
                    case "ymax":
                        y = -bbox.top
                    case "center":
                        y = -bbox.center().y
                    case _:
                        y = 0
                at = kdb.Trans(x, y)
                inst.trans = at * inst.trans
                bbox = inst.bbox()
                xmin[i_x] = min(xmin.get(i_x, None) or bbox.left, bbox.left - spacing_x)
                xmax[i_x] = max(xmax.get(i_x, None) or bbox.right, bbox.right)
                ymin[i_y] = min(
                    ymin.get(i_y, None) or bbox.bottom, bbox.bottom - spacing_y
                )
                ymax[i_y] = max(ymax.get(i_y, None) or bbox.top, bbox.top)

        insts = []
        for _ in range(shape[0]):
            insts.append([None] * shape[1])

        for i, inst in enumerate(_insts):
            i_x = i % shape[1]
            i_y = i // shape[1]
            if i_x == 0:
                y0 -= ymin.get(i_y, 0)
                x0 = 0
            else:
                x0 -= xmin.get(i_x, 0)

            if inst is not None:
                at = kdb.Trans(x0, y0)
                inst.transform(target_trans * at)
                insts[i_y][i_x] = inst
            if i_x == shape[1] - 1:
                y0 += ymax.get(i_y, 0)
                x0 = 0
            else:
                x0 += xmax.get(i_x, 0)
        return insts


def grid(
    target: KCell,
    kcells: Sequence[KCell | None] | Sequence[Sequence[KCell | None]],
    spacing: float | tuple[float, float],
    target_trans: kdb.DCplxTrans = kdb.DCplxTrans(),
    shape: tuple[int, int] | None = None,
    align_x: Literal["origin", "xmin", "xmax", "center"] = "center",
    align_y: Literal["origin", "ymin", "ymax", "center"] = "center",
    rotation: Literal[0, 1, 2, 3] = 0,
    mirror: bool = False,
) -> list[list[Instance | None]]:
    """Create a grid of instances.

    A grid uses the bounding box of the biggest width and biggest height of any bounding
    boxes inserted into the grid.
    to this bounding box a spacing is applied in x and y.

    ```
                          spacing[0] or spacing
                         ◄─►
      ┌──────────────────┐ ┌────┬─────┬─────┐ ┌────────────────┐ ┌──────────────────┐ ▲
      │                  │ │    │     │     │ │                │ │                  │ │
      │   ┌────┐         │ │    │     │     │ │                │ │      ┌──────┐    │ │
      │   │    │         │ │    │     │     │ │                │ │      │      │    │ │
      │   │    │         │ │    │ big │     │ │    ┌───────┐   │ │      │      │    │ │
      │   │    │         │ │    │ comp│     │ │    │       │   │ │      │      │    │ │
      │   │    │         │ │    │ y   │     │ │    │       │   │ │      │      │    │ │
      │   │    │         │ │    │     │     │ │    │       │   │ │      │     max bbox y
      │   └────┘         │ │    │     │     │ │    │       │   │ │      │      │    │ │
      │                  │ │    │     │     │ │    │       │   │ │      │      │    │ │
      │                  │ │    │     │     │ │    │       │   │ │      └──────┘    │ │
      │                  │ │    │     │     │ │    │       │   │ │                  │ │
      │                  │ │    │     │     │ │    └───────┘   │ │                  │ │
      │                  │ │    │     │     │ │                │ │                  │ │
     ▲└──────────────────┘ └────┴─────┴─────┘ └────────────────┘ └──────────────────┘ ▼
     │spacing[1] or spacing
     ▼┌──────────────────┐ ┌────────────────┐ ┌────────────────┐ ┌──────────────────┐
      │                  │ │                │ │                │ │                  │
      │                  │ │                │ │ ┌────┐         │ │  ┌───┐           │
      ├──────────────────┤ │ ┌───────────┐  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      │   big comp x     │ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      ├──────────────────┤ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ │           │  │ │ │    │         │ │  │   │           │
      │                  │ │ └───────────┘  │ │ └────┘         │ │  └───┘           │
      │                  │ │                │ │                │ │                  │
      └──────────────────┘ └────────────────┘ └────────────────┘ └──────────────────┘

                                                                 ◄──────────────────►
                                                                    max bbox x
    ```

    Args:
        target: Target KCell.
        kcells: Sequence or sequence of sequence of KCells to add to the grid
        spacing: Value or tuple of value (different x/y) for spacing of the grid. [um]
        target_trans: Apply a transformation to the whole grid before placing it.
        shape: Respace the input of kcells into an array and fill as many positions
            (first x then y).
        align_x: Align all the instance on the x-coordinate.
        align_y: Align all the instance on the y-coordinate.
        rotation: Apply a rotation to each kcells instance before adding it to the grid.
        mirror: Mirror the instances before placing them in the grid

    """
    if isinstance(spacing, tuple):
        spacing_x, spacing_y = spacing
    else:
        spacing_x = spacing
        spacing_y = spacing

    insts: list[list[Instance | None]]
    kcell_array: Sequence[Sequence[KCell]]

    if shape is None:
        if isinstance(kcells[0], KCell):
            kcell_array = [list(kcells)]  # type:ignore[arg-type]
        else:
            kcell_array = kcells  # type: ignore[assignment]

        x0: float = 0
        y0: float = 0

        insts = [
            [
                target.create_inst(kcell, kdb.ICplxTrans(1, rotation, mirror, 0, 0))
                for kcell in array
            ]
            for array in kcell_array
        ]
        bboxes = [
            [None if inst is None else inst.dbbox() for inst in array]
            for array in insts
        ]
        w: float = max(
            max(0 if bbox is None else bbox.width() + spacing_x for bbox in box_array)
            for box_array in bboxes
        )
        h: float = max(
            max(0 if bbox is None else bbox.height() + spacing_y for bbox in box_array)
            for box_array in bboxes
        )
        for array, bbox_array in zip(insts, bboxes):
            y0 += h - h / 2
            for bbox, inst in zip(bbox_array, array):
                x0 += w - w / 2
                if bbox is not None and inst is not None:
                    match align_x:
                        case "xmin":
                            x = -bbox.left
                        case "xmax":
                            x = -bbox.right
                        case "center":
                            x = -bbox.center().x
                        case _:
                            x = 0
                    match align_y:
                        case "ymin":
                            y = -bbox.bottom
                        case "ymax":
                            y = -bbox.top
                        case "center":
                            y = -bbox.center().y
                        case _:
                            y = 0
                    at = kdb.DCplxTrans(x0 + x, y0 + y)

                    inst.transform(target_trans * at)
                x0 += w / 2
            y0 += h / 2
            x0 = 0
        return insts
    else:
        _kcells: list[KCell | None]
        if isinstance(kcells[0], KCell):
            _kcells = kcells  # type:ignore[assignment]
        else:
            _kcells = [kcell for array in kcells for kcell in array]  # type: ignore[union-attr]

        if len(_kcells) > shape[0] * shape[1]:
            raise ValueError(
                f"Shape container size {shape[0] * shape[1]=!r} must be bigger "
                f"than the number of kcells {len(_kcells)}"
            )

        x0 = 0
        y0 = 0

        _insts = [
            None
            if kcell is None
            else target.create_inst(kcell, kdb.ICplxTrans(1, rotation, mirror, 0, 0))
            for kcell in _kcells
        ]

        insts = []
        for _ in range(shape[0]):
            insts.append([None] * shape[1])

        shape_bboxes = [None if inst is None else inst.dbbox() for inst in _insts]
        shape_bboxes_heights = [
            0 if box is None else box.height() for box in shape_bboxes
        ]
        shape_bboxes_widths = [
            0 if box is None else box.width() for box in shape_bboxes
        ]
        w = max(shape_bboxes_widths) + spacing_x
        h = max(shape_bboxes_heights) + spacing_y
        for i, (inst, bbox) in enumerate(zip(_insts, shape_bboxes)):
            i_x = i % shape[1]
            i_y = i // shape[1]
            insts[i_y][i_x] = inst
            if i_x == 0:
                y0 += h - h / 2
                x0 = 0
            else:
                x0 += w - w / 2

            if bbox is not None and inst is not None:
                match align_x:
                    case "xmin":
                        x = -bbox.left
                    case "xmax":
                        x = -bbox.right
                    case "center":
                        x = -bbox.center().x
                    case _:
                        x = 0
                match align_y:
                    case "ymin":
                        y = -bbox.bottom
                    case "ymax":
                        y = -bbox.top
                    case "center":
                        y = -bbox.center().y
                    case _:
                        y = 0
                at = kdb.DCplxTrans(x0 + x, y0 + y)

                inst.transform(target_trans * at)
            if i_x == shape[1] - 1:
                y0 += h / 2
                x0 = 0
            else:
                x0 += w / 2
        return insts


def flexgrid(
    target: KCell,
    kcells: Sequence[KCell | None] | Sequence[Sequence[KCell | None]],
    spacing: float | tuple[float, float],
    target_trans: kdb.DCplxTrans = kdb.DCplxTrans(),
    shape: tuple[int, int] | None = None,
    align_x: Literal["origin", "xmin", "xmax", "center"] = "center",
    align_y: Literal["origin", "ymin", "ymax", "center"] = "center",
    rotation: Literal[0, 1, 2, 3] = 0,
    mirror: bool = False,
) -> list[list[Instance | None]]:
    """Create a grid of instances.

    A grid uses the bounding box of the biggest width per column and biggest height per
    row of any bounding boxes inserted into the grid.
    To this bounding box a spacing is applied in x and y.

    ```
                         spacing[0] or spacing
                        ◄─►
     ┌──────────────────┐ ┌──┬─────┬──┐ ┌──────────┐ ┌──────────┐▲
     │                  │ │  │     │  │ │          │ │          ││
     │   ┌────┐         │ │  │     │  │ │          │ │   ┌──────┼│
     │   │    │         │ │  │     │  │ │          │ │   │      ││
     │   │    │         │ │  │ big │  │ │  ┌───────┤ │   │      ││
     │   │ ▼  │         │ │  │ comp│  │ │  │       │ │   │      ││
     │   │    │         │ │  │ y   │  │ │  │       │ │   │  ▼   ││
     │   │    │         │ │  │     │  │ │  │  ▼    │ │   │      ││ y[1]
     │   └────┘         │ │  │ ▼   │  │ │  │       │ │   │      ││
     │                  │ │  │     │  │ │  │       │ │   │      ││
     │                  │ │  │     │  │ │  │       │ │   └──────┼│
     │                  │ │  │     │  │ │  │       │ │          ││
     │                  │ │  │     │  │ │  └───────┤ │          ││
     │                  │ │  │     │  │ │          │ │          ││
    ▲└──────────────────┘ └──┴─────┴──┘ └──────────┘ └──────────┘▼
    │spacing[1] or spacing
    ▼┌──────────────────┐ ┌───────────┐ ┌────┬─────┐ ┌───┬──────┐▲
     ├──────────────────┤ ├───────────┤ │    │     │ │   │      ││
     │                  │ │           │ │    │     │ │   │      ││
     │              ▼   │ │           │ │    │     │ │   │      ││
     │   big comp x     │ │           │ │    │     │ │   │      ││
     │                  │ │    ▼      │ │  ▼ │     │ │ ▼ │      ││ y[0]
     ├──────────────────┤ │           │ │    │     │ │   │      ││
     │                  │ │           │ │    │     │ │   │      ││
     │                  │ │           │ │    │     │ │   │      ││
     │                  │ │           │ │    │     │ │   │      ││
     └──────────────────┴─┴───────────┴─┴────┼─────┴─┴───┼──────┘►
     ►──────────────────► ►───────────► ►──────────► ►──────────►
             x[0]               x[1]        x[2]         x[3]
    ```

    Args:
        target: Target KCell.
        kcells: Sequence or sequence of sequence of KCells to add to the grid
        spacing: Value or tuple of value (different x/y) for spacing of the grid.
        target_trans: Apply a transformation to the whole grid before placing it.
        shape: Respace the input of kcells into an array and fill as many positions
            (first x then y).
        align_x: Align all the instance on the x-coordinate.
        align_y: Align all the instance on the y-coordinate.
        rotation: Apply a rotation to each kcells instance before adding it to the grid.
        mirror: Mirror the instances before placing them in the grid

    """
    if isinstance(spacing, tuple):
        spacing_x, spacing_y = spacing
    else:
        spacing_x = spacing
        spacing_y = spacing

    insts: list[list[Instance | None]]
    kcell_array: Sequence[Sequence[KCell]]

    if shape is None:
        if isinstance(kcells[0], KCell):
            kcell_array = cast(Sequence[list[KCell]], [list(kcells)])
        else:
            kcell_array = cast(Sequence[Sequence[KCell]], kcells)

        x0: float = 0
        y0: float = 0

        insts = [
            [
                None
                if kcell is None
                else target.create_inst(
                    kcell, kdb.ICplxTrans(1, rotation, mirror, 0, 0)
                )
                for kcell in array
            ]
            for array in kcell_array
        ]
        bboxes = [
            [None if inst is None else inst.dbbox() for inst in array]
            for array in insts
        ]
        xmin: dict[float, float] = {}
        ymin: dict[float, float] = {}
        ymax: dict[float, float] = {}
        xmax: dict[float, float] = {}
        for i_y, (array, box_array) in enumerate(zip(insts, bboxes)):
            for i_x, (inst, bbox) in enumerate(zip(array, box_array)):
                if inst is not None and bbox is not None:
                    if inst is not None and bbox is not None:
                        match align_x:
                            case "xmin":
                                x = -bbox.left
                            case "xmax":
                                x = -bbox.right
                            case "center":
                                x = -bbox.center().x
                            case _:
                                x = 0
                        match align_y:
                            case "ymin":
                                y = -bbox.bottom
                            case "ymax":
                                y = -bbox.top
                            case "center":
                                y = -bbox.center().y
                            case _:
                                y = 0
                        at = kdb.DCplxTrans(x, y)
                        inst.dcplx_trans = at * inst.dcplx_trans
                        bbox = inst.dbbox()
                        xmin[i_x] = min(
                            xmin.get(i_x, None) or bbox.left, bbox.left - spacing_x
                        )
                        xmax[i_x] = max(xmax.get(i_x, None) or bbox.right, bbox.right)
                        ymin[i_y] = min(
                            ymin.get(i_y, None) or bbox.bottom, bbox.bottom - spacing_y
                        )
                        ymax[i_y] = max(ymax.get(i_y, None) or bbox.top, bbox.top)

        for i_y, (array, bbox_array) in enumerate(zip(insts, bboxes)):
            y0 -= ymin.get(i_y, 0)
            for i_x, (bbox, inst) in enumerate(zip(bbox_array, array)):
                x0 -= xmin.get(i_x, 0)
                if inst is not None and bbox is not None:
                    at = kdb.DCplxTrans(x0, y0)
                    inst.transform(target_trans * at)
                x0 += xmax.get(i_x, 0)
            y0 += ymax.get(i_y, 0)
            x0 = 0
        return insts
    else:
        _kcells: list[KCell | None]
        if isinstance(kcells[0], KCell):
            _kcells = kcells  # type:ignore[assignment]
        else:
            _kcells = [kcell for array in kcells for kcell in array]  # type: ignore[union-attr]

        if len(_kcells) > shape[0] * shape[1]:
            raise ValueError(
                f"Shape container size {shape[0] * shape[1]=} must be bigger "
                f"than the number of kcells {len(_kcells)}"
            )

        x0 = 0
        y0 = 0

        _insts = [
            None
            if kcell is None
            else target.create_inst(kcell, kdb.ICplxTrans(1, rotation, mirror, 0, 0))
            for kcell in _kcells
        ]

        xmin = {}
        ymin = {}
        ymax = {}
        xmax = {}
        for i, inst in enumerate(_insts):
            i_x = i % shape[1]
            i_y = i // shape[1]

            if inst is not None:
                bbox = inst.dbbox()
                match align_x:
                    case "xmin":
                        x = -bbox.left
                    case "xmax":
                        x = -bbox.right
                    case "center":
                        x = -bbox.center().x
                    case _:
                        x = 0
                match align_y:
                    case "ymin":
                        y = -bbox.bottom
                    case "ymax":
                        y = -bbox.top
                    case "center":
                        y = -bbox.center().y
                    case _:
                        y = 0
                at = kdb.DCplxTrans(x, y)
                inst.dcplx_trans = at * inst.dcplx_trans
                bbox = inst.dbbox()
                xmin[i_x] = min(xmin.get(i_x, None) or bbox.left, bbox.left - spacing_x)
                xmax[i_x] = max(xmax.get(i_x, None) or bbox.right, bbox.right)
                ymin[i_y] = min(
                    ymin.get(i_y, None) or bbox.bottom, bbox.bottom - spacing_y
                )
                ymax[i_y] = max(ymax.get(i_y, None) or bbox.top, bbox.top)

        insts = []
        for _ in range(shape[0]):
            insts.append([None] * shape[1])

        for i, inst in enumerate(_insts):
            i_x = i % shape[1]
            i_y = i // shape[1]
            if i_x == 0:
                y0 -= ymin.get(i_y, 0)
                x0 = 0
            else:
                x0 -= xmin.get(i_x, 0)

            if inst is not None:
                at = kdb.DCplxTrans(x0, y0)
                inst.transform(target_trans * at)
                insts[i_y][i_x] = inst
            if i_x == shape[1] - 1:
                y0 += ymax.get(i_y, 0)
                x0 = 0
            else:
                x0 += xmax.get(i_x, 0)
        return insts



"""Utilities for Ports.

Mainly renaming functions
"""

import re
from collections.abc import Callable, Iterable
from enum import IntEnum
from typing import TYPE_CHECKING, Any

from . import kdb

if TYPE_CHECKING:
    from .kcell import KCell, LayerEnum, Port


class DIRECTION(IntEnum):
    """Alias for KLayout direction to compass directions."""

    E = 0
    N = 1
    W = 2
    S = 3


def autorename(
    c: "KCell",
    f: Callable[..., None],
    *args: Any,
    **kwargs: Any,
) -> None:
    """Rename a KCell with a renaming function.

    Args:
        c: KCell to be renamed.
        f: Renaming function.
        args: Arguments for the renaming function.
        kwargs: Keyword arguments for the renaming function.
    """
    f(c.ports._ports, *args, **kwargs)


def rename_clockwise(
    ports: "Iterable[Port]",
    layer: "LayerEnum | int | None" = None,
    port_type: str | None = None,
    regex: str | None = None,
    prefix: str = "o",
    start: int = 1,
) -> None:
    """Sort and return ports in the clockwise direction.

    Args:
        ports: List of ports to rename.
        layer: Layer index / LayerEnum of port layer.
        port_type: Port type to filter the ports by.
        regex: Regex string to filter the port names by.
        prefix: Prefix to add to all ports.
        start: Start index per orientation.

    ```
             o3  o4
             |___|_
        o2 -|      |- o5
            |      |
        o1 -|______|- o6
             |   |
            o8  o7
    ```
    """
    _ports = filter_layer_pt_reg(ports, layer, port_type, regex)

    def sort_key(port: "Port") -> tuple[int, int, int]:
        match port.angle:
            case 2:
                angle = 0
            case 1:
                angle = 1
            case 0:
                angle = 2
            case 3:
                angle = 3
        dir_1 = 1 if angle < 2 else -1
        dir_2 = -1 if port.angle < 2 else 1
        key_1 = dir_1 * (
            port.trans.disp.x if angle % 2 else port.trans.disp.y
        )  # order should be y, x, -y, -x
        key_2 = dir_2 * (
            port.trans.disp.y if angle % 2 else port.trans.disp.x
        )  # order should be x, -y, -x, y

        return angle, key_1, key_2

    for i, p in enumerate(sorted(_ports, key=sort_key), start=start):
        p.name = f"{prefix}{i}"


def rename_clockwise_multi(
    ports: "Iterable[Port]",
    layers: "Iterable[LayerEnum | int] | None" = None,
    regex: str | None = None,
    type_prefix_mapping: dict[str, str] = {"optical": "o", "electrical": "e"},
    start: int = 1,
) -> None:
    """Sort and return ports in the clockwise direction.

    Args:
        ports: List of ports to rename.
        layers: Layer indexes / LayerEnums of port layers to rename.
        type_prefix_mapping: Port type to prefix matching in a dictionary.
        regex: Regex string to filter the port names by.
        start: Start index per orientation.

    ```
             o3  o4
             |___|_
        o2 -|      |- o5
            |      |
        o1 -|______|- o6
             |   |
            o8  o7
    ```
    """
    if layers:
        for p_type, prefix in type_prefix_mapping.items():
            for layer in layers:
                rename_clockwise(
                    ports=ports,
                    layer=layer,
                    port_type=p_type,
                    regex=regex,
                    prefix=prefix,
                    start=start,
                )
    else:
        for p_type, prefix in type_prefix_mapping.items():
            rename_clockwise(
                ports=ports,
                layer=None,
                port_type=p_type,
                regex=regex,
                prefix=prefix,
                start=start,
            )


def rename_by_direction(
    ports: "Iterable[Port]",
    layer: "LayerEnum | int | None" = None,
    port_type: str | None = None,
    regex: str | None = None,
    dir_names: tuple[str, str, str, str] = ("E", "N", "W", "S"),
    prefix: str = "",
) -> None:
    """Rename ports by angle of their transformation.

    Args:
        ports: list of ports to be renamed
        layer: A layer index to filter by
        port_type: port_type string to filter by
        regex: Regex string to use to filter the ports to be renamed.
        dir_names: Prefixes for the directions (east, north, west, south).
        prefix: Prefix to add before `dir_names`

    ```
             N0  N1
             |___|_
        W1 -|      |- E1
            |      |
        W0 -|______|- E0
             |   |
            S0   S1
    ```
    """
    for dir in DIRECTION:
        _ports = filter_layer_pt_reg(ports, layer, port_type, regex)
        dir_2 = -1 if dir < 2 else 1
        if dir % 2:

            def key_sort(port: "Port") -> tuple[int, int]:
                return (port.trans.disp.x, dir_2 * port.trans.disp.y)

        else:

            def key_sort(port: "Port") -> tuple[int, int]:
                return (port.trans.disp.y, dir_2 * port.trans.disp.x)

        for i, p in enumerate(sorted(filter_direction(_ports, dir), key=key_sort)):
            p.name = f"{prefix}{dir_names[dir]}{i}"


def filter_layer_pt_reg(
    ports: "Iterable[Port]",
    layer: "LayerEnum | int | None" = None,
    port_type: str | None = None,
    regex: str | None = None,
) -> "Iterable[Port]":
    """Filter ports by layer index, port type and name regex."""
    _ports = ports
    if layer is not None:
        _ports = filter_layer(_ports, layer)
    if port_type is not None:
        _ports = filter_port_type(_ports, port_type)
    if regex is not None:
        _ports = filter_regex(_ports, regex)

    return _ports


def filter_direction(ports: "Iterable[Port]", direction: int) -> "filter[Port]":
    """Filter iterable/sequence of ports by direction :py:class:~`DIRECTION`."""

    def f_func(p: "Port") -> bool:
        return p.trans.angle == direction

    return filter(f_func, ports)


def filter_orientation(ports: "Iterable[Port]", orientation: float) -> "filter[Port]":
    """Filter iterable/sequence of ports by direction :py:class:~`DIRECTION`."""

    def f_func(p: "Port") -> bool:
        return p.dcplx_trans.angle == orientation

    return filter(f_func, ports)


def filter_port_type(ports: "Iterable[Port]", port_type: str) -> "filter[Port]":
    """Filter iterable/sequence of ports by port_type."""

    def pt_filter(p: "Port") -> bool:
        return p.port_type == port_type

    return filter(pt_filter, ports)


def filter_layer(ports: "Iterable[Port]", layer: "int | LayerEnum") -> "filter[Port]":
    """Filter iterable/sequence of ports by layer index / LayerEnum."""

    def layer_filter(p: "Port") -> bool:
        return p.layer == layer

    return filter(layer_filter, ports)


def filter_regex(ports: "Iterable[Port]", regex: str) -> "filter[Port]":
    """Filter iterable/sequence of ports by port name."""
    pattern = re.compile(regex)

    def regex_filter(p: "Port") -> bool:
        if p.name is not None:
            return bool(pattern.match(p.name))
        else:
            return False

    return filter(regex_filter, ports)


polygon_dict: dict[int, kdb.Polygon] = {}


def port_polygon(width: int) -> kdb.Polygon:
    """Gets a polygon representation for a given port width."""
    if width in polygon_dict:
        return polygon_dict[width]
    else:
        poly = kdb.Polygon(
            [
                kdb.Point(0, width // 2),
                kdb.Point(0, -width // 2),
                kdb.Point(width // 2, 0),
            ]
        )

        hole = kdb.Region(poly).sized(-int(width * 0.05) or -1)
        hole -= kdb.Region(kdb.Box(0, 0, width // 2, -width // 2))

        poly.insert_hole(list(list(hole.each())[0].each_point_hull()))
        polygon_dict[width] = poly
        return poly
